/* valadatatype.vala
 *
 * Copyright (C) 2006-2008  Jürg Billeter, Raffaele Sandrini
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
 * Author:
 * 	Jürg Billeter <j@bitron.ch>
 *	Raffaele Sandrini <raffaele@sandrini.ch>
 */

#include <vala/valadatatype.h>
#include <gee/arraylist.h>
#include <gee/collection.h>
#include <gee/readonlylist.h>
#include <vala/valatypeparameter.h>
#include <vala/valacodevisitor.h>
#include <vala/valareport.h>
#include <vala/valasourcereference.h>
#include <vala/valaarraytype.h>
#include <vala/valasymbol.h>
#include <vala/valascope.h>
#include <vala/valadelegatetype.h>
#include <vala/valapointertype.h>
#include <vala/valaenum.h>
#include <vala/valastruct.h>
#include <vala/valasemanticanalyzer.h>
#include <vala/valavaluetype.h>




struct _ValaDataTypePrivate {
	gboolean _value_owned;
	gboolean _nullable;
	gboolean _is_type_argument;
	ValaTypeSymbol* _data_type;
	ValaTypeParameter* _type_parameter;
	gboolean _floating_reference;
	gboolean _is_dynamic;
	GeeList* type_argument_list;
};

#define VALA_DATA_TYPE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), VALA_TYPE_DATA_TYPE, ValaDataTypePrivate))
enum  {
	VALA_DATA_TYPE_DUMMY_PROPERTY,
	VALA_DATA_TYPE_VALUE_OWNED,
	VALA_DATA_TYPE_NULLABLE,
	VALA_DATA_TYPE_IS_TYPE_ARGUMENT,
	VALA_DATA_TYPE_DATA_TYPE,
	VALA_DATA_TYPE_TYPE_PARAMETER,
	VALA_DATA_TYPE_FLOATING_REFERENCE,
	VALA_DATA_TYPE_IS_DYNAMIC
};
static GeeList* vala_data_type__empty_type_list;
static void vala_data_type_real_accept (ValaCodeNode* base, ValaCodeVisitor* visitor);
static char* vala_data_type_real_get_cname (ValaDataType* self);
static char* vala_data_type_real_get_lower_case_cname (ValaDataType* self, const char* infix);
static char* vala_data_type_real_to_string (ValaCodeNode* base);
static char* vala_data_type_real_to_qualified_string (ValaDataType* self, ValaScope* scope);
static ValaDataType* vala_data_type_real_copy (ValaDataType* self);
static gboolean vala_data_type_real_equals (ValaDataType* self, ValaDataType* type2);
static gboolean vala_data_type_real_stricter (ValaDataType* self, ValaDataType* type2);
static void vala_data_type_real_replace_type (ValaCodeNode* base, ValaDataType* old_type, ValaDataType* new_type);
static gboolean vala_data_type_real_compatible (ValaDataType* self, ValaDataType* target_type);
static gboolean vala_data_type_real_is_invokable (ValaDataType* self);
static ValaDataType* vala_data_type_real_get_return_type (ValaDataType* self);
static GeeList* vala_data_type_real_get_parameters (ValaDataType* self);
static gboolean vala_data_type_real_is_reference_type_or_type_parameter (ValaDataType* self);
static gboolean vala_data_type_real_is_array (ValaDataType* self);
static GeeList* vala_data_type_real_get_symbols (ValaDataType* self);
static ValaSymbol* vala_data_type_real_get_member (ValaDataType* self, const char* member_name);
static ValaSymbol* vala_data_type_real_get_pointer_member (ValaDataType* self, const char* member_name);
static gboolean vala_data_type_real_is_real_struct_type (ValaDataType* self);
static char* vala_data_type_real_get_type_id (ValaDataType* self);
static char* vala_data_type_real_get_type_signature (ValaDataType* self);
static gboolean vala_data_type_real_is_disposable (ValaDataType* self);
static gpointer vala_data_type_parent_class = NULL;
static void vala_data_type_finalize (GObject * obj);



/**
 * Appends the specified type as generic type argument.
 *
 * @param arg a type reference
 */
void vala_data_type_add_type_argument (ValaDataType* self, ValaDataType* arg) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	g_return_if_fail (VALA_IS_DATA_TYPE (arg));
	if (self->priv->type_argument_list == NULL) {
		GeeList* _tmp0;
		_tmp0 = NULL;
		self->priv->type_argument_list = (_tmp0 = GEE_LIST (gee_array_list_new (VALA_TYPE_DATA_TYPE, ((GBoxedCopyFunc) (g_object_ref)), g_object_unref, g_direct_equal)), (self->priv->type_argument_list == NULL ? NULL : (self->priv->type_argument_list = (g_object_unref (self->priv->type_argument_list), NULL))), _tmp0);
	}
	gee_collection_add (GEE_COLLECTION (self->priv->type_argument_list), arg);
	vala_code_node_set_parent_node (VALA_CODE_NODE (arg), VALA_CODE_NODE (self));
}


/**
 * Returns a copy of the list of generic type arguments.
 *
 * @return type argument list
 */
GeeList* vala_data_type_get_type_arguments (ValaDataType* self) {
	GeeList* _tmp4;
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	if (self->priv->type_argument_list != NULL) {
		GeeList* _tmp0;
		_tmp0 = NULL;
		return (_tmp0 = self->priv->type_argument_list, (_tmp0 == NULL ? NULL : g_object_ref (_tmp0)));
	}
	if (vala_data_type__empty_type_list == NULL) {
		GeeList* _tmp3;
		GeeArrayList* _tmp2;
		_tmp3 = NULL;
		_tmp2 = NULL;
		vala_data_type__empty_type_list = (_tmp3 = GEE_LIST (gee_read_only_list_new (VALA_TYPE_DATA_TYPE, ((GBoxedCopyFunc) (g_object_ref)), g_object_unref, GEE_LIST ((_tmp2 = gee_array_list_new (VALA_TYPE_DATA_TYPE, ((GBoxedCopyFunc) (g_object_ref)), g_object_unref, g_direct_equal))))), (vala_data_type__empty_type_list == NULL ? NULL : (vala_data_type__empty_type_list = (g_object_unref (vala_data_type__empty_type_list), NULL))), _tmp3);
		(_tmp2 == NULL ? NULL : (_tmp2 = (g_object_unref (_tmp2), NULL)));
	}
	_tmp4 = NULL;
	return (_tmp4 = vala_data_type__empty_type_list, (_tmp4 == NULL ? NULL : g_object_ref (_tmp4)));
}


/**
 * Removes all generic type arguments.
 */
void vala_data_type_remove_all_type_arguments (ValaDataType* self) {
	GeeList* _tmp0;
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	_tmp0 = NULL;
	self->priv->type_argument_list = (_tmp0 = NULL, (self->priv->type_argument_list == NULL ? NULL : (self->priv->type_argument_list = (g_object_unref (self->priv->type_argument_list), NULL))), _tmp0);
}


static void vala_data_type_real_accept (ValaCodeNode* base, ValaCodeVisitor* visitor) {
	ValaDataType * self;
	self = VALA_DATA_TYPE (base);
	g_return_if_fail (VALA_IS_CODE_VISITOR (visitor));
	if (self->priv->type_argument_list != NULL && gee_collection_get_size (GEE_COLLECTION (self->priv->type_argument_list)) > 0) {
		{
			GeeList* type_arg_collection;
			int type_arg_it;
			type_arg_collection = self->priv->type_argument_list;
			for (type_arg_it = 0; type_arg_it < gee_collection_get_size (GEE_COLLECTION (type_arg_collection)); type_arg_it = type_arg_it + 1) {
				ValaDataType* type_arg;
				type_arg = ((ValaDataType*) (gee_list_get (GEE_LIST (type_arg_collection), type_arg_it)));
				{
					vala_code_node_accept (VALA_CODE_NODE (type_arg), visitor);
					(type_arg == NULL ? NULL : (type_arg = (g_object_unref (type_arg), NULL)));
				}
			}
		}
	}
	vala_code_visitor_visit_data_type (visitor, self);
}


/**
 * Returns the name and qualifiers of this type as it is used in C code.
 *
 * @return the type string to be used in C code
 */
static char* vala_data_type_real_get_cname (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	/* raise error*/
	vala_report_error (vala_code_node_get_source_reference (VALA_CODE_NODE (self)), "unresolved type reference");
	return NULL;
}


char* vala_data_type_get_cname (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_cname (self);
}


/**
 * Returns the name and qualifiers of this type as it is used in C code
 * in a const declaration.
 *
 * @return the type string to be used in C code const declarations
 */
char* vala_data_type_get_const_cname (ValaDataType* self) {
	char* ptr;
	ValaTypeSymbol* t;
	char* _tmp6;
	char* _tmp7;
	char* _tmp8;
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	ptr = NULL;
	t = NULL;
	/* FIXME: workaround to make constant arrays possible*/
	if (VALA_IS_ARRAY_TYPE (self)) {
		ValaTypeSymbol* _tmp1;
		ValaTypeSymbol* _tmp0;
		_tmp1 = NULL;
		_tmp0 = NULL;
		t = (_tmp1 = (_tmp0 = vala_array_type_get_element_type ((VALA_ARRAY_TYPE (self)))->priv->_data_type, (_tmp0 == NULL ? NULL : g_object_ref (_tmp0))), (t == NULL ? NULL : (t = (g_object_unref (t), NULL))), _tmp1);
	} else {
		ValaTypeSymbol* _tmp3;
		ValaTypeSymbol* _tmp2;
		_tmp3 = NULL;
		_tmp2 = NULL;
		t = (_tmp3 = (_tmp2 = self->priv->_data_type, (_tmp2 == NULL ? NULL : g_object_ref (_tmp2))), (t == NULL ? NULL : (t = (g_object_unref (t), NULL))), _tmp3);
	}
	if (!vala_typesymbol_is_reference_type (t)) {
		char* _tmp4;
		_tmp4 = NULL;
		ptr = (_tmp4 = g_strdup (""), (ptr = (g_free (ptr), NULL)), _tmp4);
	} else {
		char* _tmp5;
		_tmp5 = NULL;
		ptr = (_tmp5 = g_strdup ("*"), (ptr = (g_free (ptr), NULL)), _tmp5);
	}
	_tmp6 = NULL;
	_tmp7 = NULL;
	_tmp8 = NULL;
	return (_tmp8 = (_tmp7 = g_strdup_printf ("const %s%s", (_tmp6 = vala_typesymbol_get_cname (t, FALSE)), ptr), (_tmp6 = (g_free (_tmp6), NULL)), _tmp7), (ptr = (g_free (ptr), NULL)), (t == NULL ? NULL : (t = (g_object_unref (t), NULL))), _tmp8);
}


/**
 * Returns the C name of this data type in lower case. Words are
 * separated by underscores.
 *
 * @param infix a string to be placed between namespace and data type
 *              name or null
 * @return      the lower case name to be used in C code
 */
static char* vala_data_type_real_get_lower_case_cname (ValaDataType* self, const char* infix) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	return vala_symbol_get_lower_case_cname (VALA_SYMBOL (self->priv->_data_type), infix);
}


char* vala_data_type_get_lower_case_cname (ValaDataType* self, const char* infix) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_lower_case_cname (self, infix);
}


static char* vala_data_type_real_to_string (ValaCodeNode* base) {
	ValaDataType * self;
	self = VALA_DATA_TYPE (base);
	return vala_data_type_to_qualified_string (self, NULL);
}


static char* vala_data_type_real_to_qualified_string (ValaDataType* self, ValaScope* scope) {
	char* s;
	GeeList* type_args;
	char* _tmp20;
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	g_return_val_if_fail (scope == NULL || VALA_IS_SCOPE (scope), NULL);
	s = NULL;
	if (self->priv->_data_type != NULL) {
		ValaSymbol* _tmp0;
		ValaSymbol* global_symbol;
		ValaSymbol* sym;
		ValaScope* _tmp3;
		ValaScope* parent_scope;
		_tmp0 = NULL;
		global_symbol = (_tmp0 = VALA_SYMBOL (self->priv->_data_type), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0)));
		while (vala_symbol_get_name (vala_symbol_get_parent_symbol (global_symbol)) != NULL) {
			ValaSymbol* _tmp2;
			ValaSymbol* _tmp1;
			_tmp2 = NULL;
			_tmp1 = NULL;
			global_symbol = (_tmp2 = (_tmp1 = vala_symbol_get_parent_symbol (global_symbol), (_tmp1 == NULL ? NULL : g_object_ref (_tmp1))), (global_symbol == NULL ? NULL : (global_symbol = (g_object_unref (global_symbol), NULL))), _tmp2);
		}
		sym = NULL;
		_tmp3 = NULL;
		parent_scope = (_tmp3 = scope, (_tmp3 == NULL ? NULL : g_object_ref (_tmp3)));
		while (sym == NULL && parent_scope != NULL) {
			ValaSymbol* _tmp4;
			ValaScope* _tmp6;
			ValaScope* _tmp5;
			_tmp4 = NULL;
			sym = (_tmp4 = vala_scope_lookup (parent_scope, vala_symbol_get_name (global_symbol)), (sym == NULL ? NULL : (sym = (g_object_unref (sym), NULL))), _tmp4);
			_tmp6 = NULL;
			_tmp5 = NULL;
			parent_scope = (_tmp6 = (_tmp5 = vala_scope_get_parent_scope (parent_scope), (_tmp5 == NULL ? NULL : g_object_ref (_tmp5))), (parent_scope == NULL ? NULL : (parent_scope = (g_object_unref (parent_scope), NULL))), _tmp6);
		}
		if (sym != NULL && global_symbol != sym) {
			char* _tmp8;
			char* _tmp7;
			_tmp8 = NULL;
			_tmp7 = NULL;
			s = (_tmp8 = g_strconcat ("global::", (_tmp7 = vala_symbol_get_full_name (VALA_SYMBOL (self->priv->_data_type))), NULL), (s = (g_free (s), NULL)), _tmp8);
			_tmp7 = (g_free (_tmp7), NULL);
			;
		} else {
			char* _tmp9;
			_tmp9 = NULL;
			s = (_tmp9 = vala_symbol_get_full_name (VALA_SYMBOL (self->priv->_data_type)), (s = (g_free (s), NULL)), _tmp9);
		}
		(global_symbol == NULL ? NULL : (global_symbol = (g_object_unref (global_symbol), NULL)));
		(sym == NULL ? NULL : (sym = (g_object_unref (sym), NULL)));
		(parent_scope == NULL ? NULL : (parent_scope = (g_object_unref (parent_scope), NULL)));
	} else {
		if (self->priv->_type_parameter != NULL) {
			char* _tmp11;
			const char* _tmp10;
			_tmp11 = NULL;
			_tmp10 = NULL;
			s = (_tmp11 = (_tmp10 = vala_symbol_get_name (VALA_SYMBOL (self->priv->_type_parameter)), (_tmp10 == NULL ? NULL : g_strdup (_tmp10))), (s = (g_free (s), NULL)), _tmp11);
		} else {
			char* _tmp12;
			_tmp12 = NULL;
			s = (_tmp12 = g_strdup ("null"), (s = (g_free (s), NULL)), _tmp12);
		}
	}
	type_args = vala_data_type_get_type_arguments (self);
	if (!(VALA_IS_ARRAY_TYPE (self)) && gee_collection_get_size (GEE_COLLECTION (type_args)) > 0) {
		char* _tmp13;
		gboolean first;
		char* _tmp18;
		_tmp13 = NULL;
		s = (_tmp13 = g_strconcat (s, ("<"), NULL), (s = (g_free (s), NULL)), _tmp13);
		first = TRUE;
		{
			GeeList* type_arg_collection;
			int type_arg_it;
			type_arg_collection = type_args;
			for (type_arg_it = 0; type_arg_it < gee_collection_get_size (GEE_COLLECTION (type_arg_collection)); type_arg_it = type_arg_it + 1) {
				ValaDataType* type_arg;
				type_arg = ((ValaDataType*) (gee_list_get (GEE_LIST (type_arg_collection), type_arg_it)));
				{
					char* _tmp17;
					char* _tmp16;
					if (!first) {
						char* _tmp14;
						_tmp14 = NULL;
						s = (_tmp14 = g_strconcat (s, (","), NULL), (s = (g_free (s), NULL)), _tmp14);
					} else {
						first = FALSE;
					}
					if (!type_arg->priv->_value_owned) {
						char* _tmp15;
						_tmp15 = NULL;
						s = (_tmp15 = g_strconcat (s, ("weak "), NULL), (s = (g_free (s), NULL)), _tmp15);
					}
					_tmp17 = NULL;
					_tmp16 = NULL;
					s = (_tmp17 = g_strconcat (s, (_tmp16 = (vala_data_type_to_qualified_string (type_arg, scope))), NULL), (s = (g_free (s), NULL)), _tmp17);
					_tmp16 = (g_free (_tmp16), NULL);
					(type_arg == NULL ? NULL : (type_arg = (g_object_unref (type_arg), NULL)));
				}
			}
		}
		_tmp18 = NULL;
		s = (_tmp18 = g_strconcat (s, (">"), NULL), (s = (g_free (s), NULL)), _tmp18);
	}
	if (self->priv->_nullable) {
		char* _tmp19;
		_tmp19 = NULL;
		s = (_tmp19 = g_strconcat (s, ("?"), NULL), (s = (g_free (s), NULL)), _tmp19);
	}
	_tmp20 = NULL;
	return (_tmp20 = s, (type_args == NULL ? NULL : (type_args = (g_object_unref (type_args), NULL))), _tmp20);
}


char* vala_data_type_to_qualified_string (ValaDataType* self, ValaScope* scope) {
	return VALA_DATA_TYPE_GET_CLASS (self)->to_qualified_string (self, scope);
}


static ValaDataType* vala_data_type_real_copy (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	g_critical ("Type `%s' does not implement abstract method `vala_data_type_copy'", g_type_name (G_TYPE_FROM_INSTANCE (self)));
	return NULL;
}


/**
 * Creates a shallow copy of this type reference.
 *
 * @return copy of this type reference
 */
ValaDataType* vala_data_type_copy (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->copy (self);
}


/**
 * Checks two type references for equality. May only be used with
 * resolved type references.
 *
 * @param type2 a type reference
 * @return      true if this type reference is equal to type2, false
 *              otherwise
 */
static gboolean vala_data_type_real_equals (ValaDataType* self, ValaDataType* type2) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	g_return_val_if_fail (VALA_IS_DATA_TYPE (type2), FALSE);
	if (type2->priv->_value_owned != self->priv->_value_owned) {
		return FALSE;
	}
	if (type2->priv->_nullable != self->priv->_nullable) {
		return FALSE;
	}
	if (type2->priv->_data_type != self->priv->_data_type) {
		return FALSE;
	}
	if (type2->priv->_type_parameter != NULL || self->priv->_type_parameter != NULL) {
		if (type2->priv->_type_parameter == NULL || self->priv->_type_parameter == NULL) {
			return FALSE;
		}
		if (!vala_typeparameter_equals (type2->priv->_type_parameter, self->priv->_type_parameter)) {
			return FALSE;
		}
	}
	if (type2->priv->_floating_reference != self->priv->_floating_reference) {
		return FALSE;
	}
	return TRUE;
}


gboolean vala_data_type_equals (ValaDataType* self, ValaDataType* type2) {
	return VALA_DATA_TYPE_GET_CLASS (self)->equals (self, type2);
}


/**
 * Checks whether this type reference is at least as strict as the
 * specified type reference type2.
 *
 * @param type2 a type reference
 * @return      true if this type reference is stricter or equal
 */
static gboolean vala_data_type_real_stricter (ValaDataType* self, ValaDataType* type2) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	g_return_val_if_fail (VALA_IS_DATA_TYPE (type2), FALSE);
	if (type2->priv->_value_owned != self->priv->_value_owned) {
		return FALSE;
	}
	if (!type2->priv->_nullable && self->priv->_nullable) {
		return FALSE;
	}
	/* temporarily ignore type parameters */
	if (self->priv->_type_parameter != NULL || type2->priv->_type_parameter != NULL) {
		return TRUE;
	}
	if (type2->priv->_data_type != self->priv->_data_type) {
		/* FIXME: allow this type reference to refer to a
		        subtype of the type type2 is referring to*/
		return FALSE;
	}
	if (type2->priv->_floating_reference != self->priv->_floating_reference) {
		return FALSE;
	}
	return TRUE;
}


gboolean vala_data_type_stricter (ValaDataType* self, ValaDataType* type2) {
	return VALA_DATA_TYPE_GET_CLASS (self)->stricter (self, type2);
}


static void vala_data_type_real_replace_type (ValaCodeNode* base, ValaDataType* old_type, ValaDataType* new_type) {
	ValaDataType * self;
	self = VALA_DATA_TYPE (base);
	g_return_if_fail (VALA_IS_DATA_TYPE (old_type));
	g_return_if_fail (VALA_IS_DATA_TYPE (new_type));
	if (self->priv->type_argument_list != NULL) {
		{
			gint i;
			i = 0;
			for (; i < gee_collection_get_size (GEE_COLLECTION (self->priv->type_argument_list)); i++) {
				ValaDataType* _tmp0;
				gboolean _tmp1;
				_tmp0 = NULL;
				if ((_tmp1 = (_tmp0 = ((ValaDataType*) (gee_list_get (((GeeList*) (self->priv->type_argument_list)), i)))) == old_type, (_tmp0 == NULL ? NULL : (_tmp0 = (g_object_unref (_tmp0), NULL))), _tmp1)) {
					gee_list_set (((GeeList*) (self->priv->type_argument_list)), i, new_type);
					return;
				}
			}
		}
	}
}


static gboolean vala_data_type_real_compatible (ValaDataType* self, ValaDataType* target_type) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	g_return_val_if_fail (VALA_IS_DATA_TYPE (target_type), FALSE);
	if (VALA_IS_DELEGATE_TYPE (target_type) && VALA_IS_DELEGATE_TYPE (self)) {
		return vala_delegate_type_get_delegate_symbol ((VALA_DELEGATE_TYPE (target_type))) == vala_delegate_type_get_delegate_symbol ((VALA_DELEGATE_TYPE (self)));
	}
	if (VALA_IS_POINTER_TYPE (target_type)) {
		/* any reference or array type or pointer type can be cast to a generic pointer */
		if (self->priv->_type_parameter != NULL || (self->priv->_data_type != NULL && (vala_typesymbol_is_reference_type (self->priv->_data_type) || VALA_IS_DELEGATE_TYPE (self)))) {
			return TRUE;
		}
		return FALSE;
	}
	/* temporarily ignore type parameters */
	if (target_type->priv->_type_parameter != NULL) {
		return TRUE;
	}
	if (VALA_IS_ARRAY_TYPE (self) != VALA_IS_ARRAY_TYPE (target_type)) {
		return FALSE;
	}
	if (VALA_IS_ENUM (self->priv->_data_type) && VALA_IS_STRUCT (target_type->priv->_data_type) && vala_struct_is_integer_type ((VALA_STRUCT (target_type->priv->_data_type)))) {
		return TRUE;
	}
	if (self->priv->_data_type == target_type->priv->_data_type) {
		return TRUE;
	}
	if (VALA_IS_STRUCT (self->priv->_data_type) && VALA_IS_STRUCT (target_type->priv->_data_type)) {
		ValaStruct* _tmp7;
		ValaStruct* expr_struct;
		ValaStruct* _tmp8;
		ValaStruct* expect_struct;
		_tmp7 = NULL;
		expr_struct = (_tmp7 = VALA_STRUCT (self->priv->_data_type), (_tmp7 == NULL ? NULL : g_object_ref (_tmp7)));
		_tmp8 = NULL;
		expect_struct = (_tmp8 = VALA_STRUCT (target_type->priv->_data_type), (_tmp8 == NULL ? NULL : g_object_ref (_tmp8)));
		/* integer types may be implicitly cast to floating point types */
		if (vala_struct_is_integer_type (expr_struct) && vala_struct_is_floating_type (expect_struct)) {
			gboolean _tmp9;
			return (_tmp9 = TRUE, (expr_struct == NULL ? NULL : (expr_struct = (g_object_unref (expr_struct), NULL))), (expect_struct == NULL ? NULL : (expect_struct = (g_object_unref (expect_struct), NULL))), _tmp9);
		}
		if ((vala_struct_is_integer_type (expr_struct) && vala_struct_is_integer_type (expect_struct)) || (vala_struct_is_floating_type (expr_struct) && vala_struct_is_floating_type (expect_struct))) {
			if (vala_struct_get_rank (expr_struct) <= vala_struct_get_rank (expect_struct)) {
				gboolean _tmp10;
				return (_tmp10 = TRUE, (expr_struct == NULL ? NULL : (expr_struct = (g_object_unref (expr_struct), NULL))), (expect_struct == NULL ? NULL : (expect_struct = (g_object_unref (expect_struct), NULL))), _tmp10);
			}
		}
		(expr_struct == NULL ? NULL : (expr_struct = (g_object_unref (expr_struct), NULL)));
		(expect_struct == NULL ? NULL : (expect_struct = (g_object_unref (expect_struct), NULL)));
	}
	if (self->priv->_data_type != NULL && target_type->priv->_data_type != NULL && vala_typesymbol_is_subtype_of (self->priv->_data_type, target_type->priv->_data_type)) {
		return TRUE;
	}
	return FALSE;
}


gboolean vala_data_type_compatible (ValaDataType* self, ValaDataType* target_type) {
	return VALA_DATA_TYPE_GET_CLASS (self)->compatible (self, target_type);
}


/**
 * Returns whether instances of this type are invokable.
 *
 * @return true if invokable, false otherwise
 */
static gboolean vala_data_type_real_is_invokable (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return FALSE;
}


gboolean vala_data_type_is_invokable (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->is_invokable (self);
}


/**
 * Returns the return type of this invokable.
 *
 * @return return type
 */
static ValaDataType* vala_data_type_real_get_return_type (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	return NULL;
}


ValaDataType* vala_data_type_get_return_type (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_return_type (self);
}


/**
 * Returns copy of the list of invocation parameters.
 *
 * @return parameter list
 */
static GeeList* vala_data_type_real_get_parameters (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	return NULL;
}


GeeList* vala_data_type_get_parameters (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_parameters (self);
}


static gboolean vala_data_type_real_is_reference_type_or_type_parameter (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return (self->priv->_data_type != NULL && vala_typesymbol_is_reference_type (self->priv->_data_type)) || self->priv->_type_parameter != NULL;
}


gboolean vala_data_type_is_reference_type_or_type_parameter (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->is_reference_type_or_type_parameter (self);
}


static gboolean vala_data_type_real_is_array (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return FALSE;
}


gboolean vala_data_type_is_array (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->is_array (self);
}


/**
 * Returns a list of symbols that define this type.
 *
 * @return symbol list
 */
static GeeList* vala_data_type_real_get_symbols (ValaDataType* self) {
	GeeArrayList* symbols;
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	symbols = gee_array_list_new (VALA_TYPE_SYMBOL, ((GBoxedCopyFunc) (g_object_ref)), g_object_unref, g_direct_equal);
	if (self->priv->_data_type != NULL) {
		gee_collection_add (GEE_COLLECTION (symbols), VALA_SYMBOL (self->priv->_data_type));
	}
	return GEE_LIST (symbols);
}


GeeList* vala_data_type_get_symbols (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_symbols (self);
}


static ValaSymbol* vala_data_type_real_get_member (ValaDataType* self, const char* member_name) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	g_return_val_if_fail (member_name != NULL, NULL);
	if (self->priv->_data_type != NULL) {
		return vala_semantic_analyzer_symbol_lookup_inherited (VALA_SYMBOL (self->priv->_data_type), member_name);
	}
	return NULL;
}


ValaSymbol* vala_data_type_get_member (ValaDataType* self, const char* member_name) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_member (self, member_name);
}


static ValaSymbol* vala_data_type_real_get_pointer_member (ValaDataType* self, const char* member_name) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	g_return_val_if_fail (member_name != NULL, NULL);
	return NULL;
}


ValaSymbol* vala_data_type_get_pointer_member (ValaDataType* self, const char* member_name) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_pointer_member (self, member_name);
}


/**
 * Checks whether this data type references a real struct. A real struct
 * is a struct which is not a simple (fundamental) type.
 */
static gboolean vala_data_type_real_is_real_struct_type (ValaDataType* self) {
	ValaStruct* _tmp1;
	ValaTypeSymbol* _tmp0;
	ValaStruct* s;
	gboolean _tmp3;
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	_tmp1 = NULL;
	_tmp0 = NULL;
	s = (_tmp1 = (_tmp0 = self->priv->_data_type, (VALA_IS_STRUCT (_tmp0) ? ((ValaStruct*) (_tmp0)) : NULL)), (_tmp1 == NULL ? NULL : g_object_ref (_tmp1)));
	if (s != NULL && !vala_struct_is_simple_type (s)) {
		gboolean _tmp2;
		return (_tmp2 = TRUE, (s == NULL ? NULL : (s = (g_object_unref (s), NULL))), _tmp2);
	}
	return (_tmp3 = FALSE, (s == NULL ? NULL : (s = (g_object_unref (s), NULL))), _tmp3);
}


gboolean vala_data_type_is_real_struct_type (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->is_real_struct_type (self);
}


static char* vala_data_type_real_get_type_id (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	if (self->priv->_data_type != NULL) {
		return vala_typesymbol_get_type_id (self->priv->_data_type);
	} else {
		return NULL;
	}
}


char* vala_data_type_get_type_id (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_type_id (self);
}


/**
 * Returns type signature as used for GVariant and D-Bus.
 */
static char* vala_data_type_real_get_type_signature (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	if (self->priv->_data_type != NULL) {
		return vala_typesymbol_get_type_signature (self->priv->_data_type);
	} else {
		return NULL;
	}
}


char* vala_data_type_get_type_signature (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->get_type_signature (self);
}


/**
 * Returns whether the value needs to be disposed, i.e. whether
 * allocated memory or other resources need to be released when
 * the value is no longer needed.
 */
static gboolean vala_data_type_real_is_disposable (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	if (!self->priv->_value_owned) {
		return FALSE;
	}
	if (vala_data_type_is_reference_type_or_type_parameter (self)) {
		return TRUE;
	} else {
		if (VALA_IS_VALUE_TYPE (self)) {
			/* nullable structs are heap allocated*/
			return self->priv->_nullable;
		}
	}
	return FALSE;
}


gboolean vala_data_type_is_disposable (ValaDataType* self) {
	return VALA_DATA_TYPE_GET_CLASS (self)->is_disposable (self);
}


gboolean vala_data_type_get_value_owned (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return self->priv->_value_owned;
}


void vala_data_type_set_value_owned (ValaDataType* self, gboolean value) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	self->priv->_value_owned = value;
	g_object_notify (((GObject *) (self)), "value-owned");
}


gboolean vala_data_type_get_nullable (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return self->priv->_nullable;
}


void vala_data_type_set_nullable (ValaDataType* self, gboolean value) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	self->priv->_nullable = value;
	g_object_notify (((GObject *) (self)), "nullable");
}


gboolean vala_data_type_get_is_type_argument (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return self->priv->_is_type_argument;
}


void vala_data_type_set_is_type_argument (ValaDataType* self, gboolean value) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	self->priv->_is_type_argument = value;
	g_object_notify (((GObject *) (self)), "is-type-argument");
}


ValaTypeSymbol* vala_data_type_get_data_type (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	return self->priv->_data_type;
}


void vala_data_type_set_data_type (ValaDataType* self, ValaTypeSymbol* value) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	self->priv->_data_type = value;
	g_object_notify (((GObject *) (self)), "data-type");
}


ValaTypeParameter* vala_data_type_get_type_parameter (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), NULL);
	return self->priv->_type_parameter;
}


void vala_data_type_set_type_parameter (ValaDataType* self, ValaTypeParameter* value) {
	ValaTypeParameter* _tmp2;
	ValaTypeParameter* _tmp1;
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_type_parameter = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_object_ref (_tmp1))), (self->priv->_type_parameter == NULL ? NULL : (self->priv->_type_parameter = (g_object_unref (self->priv->_type_parameter), NULL))), _tmp2);
	g_object_notify (((GObject *) (self)), "type-parameter");
}


gboolean vala_data_type_get_floating_reference (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return self->priv->_floating_reference;
}


void vala_data_type_set_floating_reference (ValaDataType* self, gboolean value) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	self->priv->_floating_reference = value;
	g_object_notify (((GObject *) (self)), "floating-reference");
}


gboolean vala_data_type_get_is_dynamic (ValaDataType* self) {
	g_return_val_if_fail (VALA_IS_DATA_TYPE (self), FALSE);
	return self->priv->_is_dynamic;
}


void vala_data_type_set_is_dynamic (ValaDataType* self, gboolean value) {
	g_return_if_fail (VALA_IS_DATA_TYPE (self));
	self->priv->_is_dynamic = value;
	g_object_notify (((GObject *) (self)), "is-dynamic");
}


static void vala_data_type_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	ValaDataType * self;
	self = VALA_DATA_TYPE (object);
	switch (property_id) {
		case VALA_DATA_TYPE_VALUE_OWNED:
		g_value_set_boolean (value, vala_data_type_get_value_owned (self));
		break;
		case VALA_DATA_TYPE_NULLABLE:
		g_value_set_boolean (value, vala_data_type_get_nullable (self));
		break;
		case VALA_DATA_TYPE_IS_TYPE_ARGUMENT:
		g_value_set_boolean (value, vala_data_type_get_is_type_argument (self));
		break;
		case VALA_DATA_TYPE_DATA_TYPE:
		g_value_set_object (value, vala_data_type_get_data_type (self));
		break;
		case VALA_DATA_TYPE_TYPE_PARAMETER:
		g_value_set_object (value, vala_data_type_get_type_parameter (self));
		break;
		case VALA_DATA_TYPE_FLOATING_REFERENCE:
		g_value_set_boolean (value, vala_data_type_get_floating_reference (self));
		break;
		case VALA_DATA_TYPE_IS_DYNAMIC:
		g_value_set_boolean (value, vala_data_type_get_is_dynamic (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void vala_data_type_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	ValaDataType * self;
	self = VALA_DATA_TYPE (object);
	switch (property_id) {
		case VALA_DATA_TYPE_VALUE_OWNED:
		vala_data_type_set_value_owned (self, g_value_get_boolean (value));
		break;
		case VALA_DATA_TYPE_NULLABLE:
		vala_data_type_set_nullable (self, g_value_get_boolean (value));
		break;
		case VALA_DATA_TYPE_IS_TYPE_ARGUMENT:
		vala_data_type_set_is_type_argument (self, g_value_get_boolean (value));
		break;
		case VALA_DATA_TYPE_DATA_TYPE:
		vala_data_type_set_data_type (self, g_value_get_object (value));
		break;
		case VALA_DATA_TYPE_TYPE_PARAMETER:
		vala_data_type_set_type_parameter (self, g_value_get_object (value));
		break;
		case VALA_DATA_TYPE_FLOATING_REFERENCE:
		vala_data_type_set_floating_reference (self, g_value_get_boolean (value));
		break;
		case VALA_DATA_TYPE_IS_DYNAMIC:
		vala_data_type_set_is_dynamic (self, g_value_get_boolean (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void vala_data_type_class_init (ValaDataTypeClass * klass) {
	vala_data_type_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (ValaDataTypePrivate));
	G_OBJECT_CLASS (klass)->get_property = vala_data_type_get_property;
	G_OBJECT_CLASS (klass)->set_property = vala_data_type_set_property;
	G_OBJECT_CLASS (klass)->finalize = vala_data_type_finalize;
	VALA_CODE_NODE_CLASS (klass)->accept = vala_data_type_real_accept;
	VALA_DATA_TYPE_CLASS (klass)->get_cname = vala_data_type_real_get_cname;
	VALA_DATA_TYPE_CLASS (klass)->get_lower_case_cname = vala_data_type_real_get_lower_case_cname;
	VALA_CODE_NODE_CLASS (klass)->to_string = vala_data_type_real_to_string;
	VALA_DATA_TYPE_CLASS (klass)->to_qualified_string = vala_data_type_real_to_qualified_string;
	VALA_DATA_TYPE_CLASS (klass)->copy = vala_data_type_real_copy;
	VALA_DATA_TYPE_CLASS (klass)->equals = vala_data_type_real_equals;
	VALA_DATA_TYPE_CLASS (klass)->stricter = vala_data_type_real_stricter;
	VALA_CODE_NODE_CLASS (klass)->replace_type = vala_data_type_real_replace_type;
	VALA_DATA_TYPE_CLASS (klass)->compatible = vala_data_type_real_compatible;
	VALA_DATA_TYPE_CLASS (klass)->is_invokable = vala_data_type_real_is_invokable;
	VALA_DATA_TYPE_CLASS (klass)->get_return_type = vala_data_type_real_get_return_type;
	VALA_DATA_TYPE_CLASS (klass)->get_parameters = vala_data_type_real_get_parameters;
	VALA_DATA_TYPE_CLASS (klass)->is_reference_type_or_type_parameter = vala_data_type_real_is_reference_type_or_type_parameter;
	VALA_DATA_TYPE_CLASS (klass)->is_array = vala_data_type_real_is_array;
	VALA_DATA_TYPE_CLASS (klass)->get_symbols = vala_data_type_real_get_symbols;
	VALA_DATA_TYPE_CLASS (klass)->get_member = vala_data_type_real_get_member;
	VALA_DATA_TYPE_CLASS (klass)->get_pointer_member = vala_data_type_real_get_pointer_member;
	VALA_DATA_TYPE_CLASS (klass)->is_real_struct_type = vala_data_type_real_is_real_struct_type;
	VALA_DATA_TYPE_CLASS (klass)->get_type_id = vala_data_type_real_get_type_id;
	VALA_DATA_TYPE_CLASS (klass)->get_type_signature = vala_data_type_real_get_type_signature;
	VALA_DATA_TYPE_CLASS (klass)->is_disposable = vala_data_type_real_is_disposable;
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_VALUE_OWNED, g_param_spec_boolean ("value-owned", "value-owned", "value-owned", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_NULLABLE, g_param_spec_boolean ("nullable", "nullable", "nullable", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_IS_TYPE_ARGUMENT, g_param_spec_boolean ("is-type-argument", "is-type-argument", "is-type-argument", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_DATA_TYPE, g_param_spec_object ("data-type", "data-type", "data-type", VALA_TYPE_TYPESYMBOL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_TYPE_PARAMETER, g_param_spec_object ("type-parameter", "type-parameter", "type-parameter", VALA_TYPE_TYPEPARAMETER, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_FLOATING_REFERENCE, g_param_spec_boolean ("floating-reference", "floating-reference", "floating-reference", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), VALA_DATA_TYPE_IS_DYNAMIC, g_param_spec_boolean ("is-dynamic", "is-dynamic", "is-dynamic", FALSE, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}


static void vala_data_type_instance_init (ValaDataType * self) {
	self->priv = VALA_DATA_TYPE_GET_PRIVATE (self);
}


static void vala_data_type_finalize (GObject * obj) {
	ValaDataType * self;
	self = VALA_DATA_TYPE (obj);
	(self->priv->_type_parameter == NULL ? NULL : (self->priv->_type_parameter = (g_object_unref (self->priv->_type_parameter), NULL)));
	(self->priv->type_argument_list == NULL ? NULL : (self->priv->type_argument_list = (g_object_unref (self->priv->type_argument_list), NULL)));
	G_OBJECT_CLASS (vala_data_type_parent_class)->finalize (obj);
}


GType vala_data_type_get_type (void) {
	static GType vala_data_type_type_id = 0;
	if (vala_data_type_type_id == 0) {
		static const GTypeInfo g_define_type_info = { sizeof (ValaDataTypeClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) vala_data_type_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (ValaDataType), 0, (GInstanceInitFunc) vala_data_type_instance_init };
		vala_data_type_type_id = g_type_register_static (VALA_TYPE_CODE_NODE, "ValaDataType", &g_define_type_info, G_TYPE_FLAG_ABSTRACT);
	}
	return vala_data_type_type_id;
}




